--
-- Track Resources Addon
--

local addOnName, TR = ...
local lootProcessedCache = {}
local castExtractGasTime = 0
local FIND_MINERALS = 2580
local FIND_HERBS = 2383
local SPELL_ID_EXTRACT_GAS = 30427
local PROF_ICON_MINERALS = 4620679
local PROF_ICON_HERBS = 4620675
local PROF_ICON_ENGINEERING = 4620673
local profNames = {}
local profNames_rev = {}
local GAME_BUILD = select(4,GetBuildInfo())

TR.cache = {
    items = {}
}
TR.refreshRate = 0.3
TR.iconSettings = {
    alpha = 0.75,
    width = 12,
    height = 12,
    alphaMinimapNear = 0.3,
    alphaMinimapFar = 0.5,
}

function TrackResources:OnLoad()
    self:RegisterEvent("LOOT_READY")
    self:RegisterEvent("LOOT_CLOSED")
    self:RegisterEvent("SKILL_LINES_CHANGED")
    self:RegisterEvent("SPELLS_CHANGED")
    self:RegisterEvent("PLAYER_ENTERING_WORLD")

    self:RegisterEvent("ITEM_PUSH")
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")

    profNames = TR.ProfessionNames[GetLocale()]
    profNames_rev = tInvert(profNames)

    if not TrackResourcesCfg then
        TrackResourcesCfg = {
            settings = {
                showMinimap = true,
                showZonemap = true,
                sizeMinimap = 12,
                sizeZonemap = 12,
                alwaysShow  = false,
                filterMode  = "LINEAR",
                minDistanceToCloud = 3
            },
            version = 2,
            db = {}
        }
        self:Print("New DB Initialised")
    end

    TrackResourcesCfg.settings.filterMode =  TrackResourcesCfg.settings.filterMode or "LINEAR"
    TrackResourcesCfg.settings.alpha = TrackResourcesCfg.settings.alpha or TR.iconSettings.alpha
    TrackResourcesCfg.settings.autocast = TrackResourcesCfg.settings.autocast == nil and true or TrackResourcesCfg.settings.autocast
    TrackResourcesCfg.settings.minDistanceToCloud = TrackResourcesCfg.settings.minDistanceToCloud or 3

    -- Reset everyone to LINEAR due to visual bugs and beta builds crashing.
    -- Probably a better way of handling this but meh... this is quicker.
    if TrackResourcesCfg.version == 1 then
        if TrackResourcesCfg.settings.filterMode:upper() == "TRILINEAR" then
            self:Important("Resetting map icon texture to LINEAR mode due to gfx glitches.  If you want to return to TRILINEAR mode you can run: /tr filtermode trilinear")
            TrackResourcesCfg.settings.filterMode = "LINEAR"
        end
        TrackResourcesCfg.version = 2
    end

    SLASH_TR1 = '/tr'
    SlashCmdList["TR"] = function (text, chatBoxFrame) self:ParseCommands(text, chatBoxFrame) end

    -- Lookup item info for cache on initial load
    for objectID, itemID in pairs(TR.objectFilter) do
        self:GetItemInfo(itemID, TR.objectNameOverride[objectID])
    end
end

function TrackResources:OnEvent(event, ...)
    if self[event] then
        return self[event](self, ...)
    end
end

function TrackResources:ADDON_LOADED(loadedAddOnName)
    if loadedAddOnName == addOnName then
        self:OnLoad()
    end
end

function TrackResources:LOOT_READY(autoloot)
    self:ProcessObject()
end

function TrackResources:LOOT_CLOSED()
    lootProcessedCache = {}
end

function TrackResources:PLAYER_ENTERING_WORLD(isInitialLogin, isReloadingUi)
    self:CheckProfessionSpells()
end

function TrackResources:SPELLS_CHANGED()
    self:CheckProfessionSpells()
end

function TrackResources:SKILL_LINES_CHANGED(...)
    self:CheckProfessionSpells()
    TrackResourcesMinimap:RefreshIcons()
end

function TrackResources:UNIT_SPELLCAST_CHANNEL_STOP(unitTarget, castGUID, spellID)
    if spellID == SPELL_ID_EXTRACT_GAS then
        castExtractGasTime = time()
    end
end

function TrackResources:ITEM_PUSH(bagSlot, iconFileID)
    -- Try and keep this detection logic to shortly after Extract Gas has finished channelling
    if time() - castExtractGasTime > 2 then
        return
    end

    local itemID = TR.gasClouds[iconFileID]
    if not itemID then
        -- Debug("Unknown iconFileID %d", iconFileID)
        return
    end

    local uiMapID, x, y = self:GetPosition()
    local textureID = self:GetItemInfo(itemID, nil)
    if uiMapID and x and y and textureID then
        self:AddResourceToDB(itemID, textureID, uiMapID, x, y, nil)
    end
end

function TrackResources:CheckProfessionSpells()
    -- If before Cata (which added GetProfessions()...)
    if (GAME_BUILD < 40000) then
        TR.isEngineer = false
        for skillIndex = 1, GetNumSkillLines() do
            local skillName= GetSkillLineInfo(skillIndex)
            if profNames_rev[skillName] == 202 then
                TR.isEngineer = true
            end
        end

        TR.isHerbalist = TrackResourcesCfg.settings.alwaysShow or DoesSpellExist(select(1, GetSpellInfo(FIND_HERBS)))
        TR.isMiner = TrackResourcesCfg.settings.alwaysShow or DoesSpellExist(select(1, GetSpellInfo(FIND_MINERALS)))
    else
        local icon1 = nil
        local icon2 = nil

        local prof1, prof2 = GetProfessions()
        if prof1 then _, icon1 = GetProfessionInfo(prof1) end
        if prof2 then _, icon2 = GetProfessionInfo(prof2) end

        TR.isHerbalist = TrackResourcesCfg.settings.alwaysShow or
                         icon1 == PROF_ICON_HERBS or
                         icon2 == PROF_ICON_HERBS

        TR.isMiner = TrackResourcesCfg.settings.alwaysShow or
                     icon1 == PROF_ICON_MINERALS or
                     icon2 == PROF_ICON_MINERALS

        TR.isEngineer = TrackResourcesCfg.settings.alwaysShow or
                        icon1 == PROF_ICON_ENGINEERING or
                        icon2 == PROF_ICON_ENGINEERING

    end
end

function TrackResources:ProcessObject()
    for index = 1, GetNumLootItems() do
        if LootSlotHasItem(index) then
            local guid = GetLootSourceInfo(index)
            if lootProcessedCache[guid] then return end

            local unitType, _, _, _, _, objectID = strsplit("-", guid);
            if unitType ~= "GameObject" then return end
            objectID = tonumber(objectID)

            -- Debug("objectID: %d via guid %s", objectID, guid)
            local itemID = TR.objectFilter[objectID]
            if not itemID then
                --Debug("Missing objectID: %d via guid %s", objectID, guid)
                lootProcessedCache[guid] = true
                return
            end
            itemID = tonumber(itemID)

            local uiMapID, x, y = self:GetPosition()
            local textureID = self:GetItemInfo(itemID, TR.objectNameOverride[objectID])
            if uiMapID and x and y and textureID then
                self:AddResourceToDB(itemID, textureID, uiMapID, x, y, TR.objectNameOverride[objectID])
            end
            lootProcessedCache[guid] = true
            return
        end
    end
end

-- for extra stuff we may want to add later
function TrackResources:GetItemInfo(itemID, nameItemID)
    itemID = tonumber(itemID)
    if not TR.cache.items[itemID] then
        -- instant data we want now..
        local iconItemID = TR.iconTextureOverride[itemID] or itemID
        local textureID = select(5, GetItemInfoInstant(iconItemID))
        TR.cache.items[itemID] = { textureID = textureID }

        -- cache other stuff too
        -- todo: expand texture override to include name and rarity too and skip this if set
        local item = Item:CreateFromItemID(itemID)
        item:ContinueOnItemLoad(function()
            local itemName, _, itemRarity = GetItemInfo(itemID)
            TR.cache.items[itemID].itemName = itemName
            TR.cache.items[itemID].itemRarity = itemRarity
        end)
    end

    if nameItemID and nameItemID > 0 then
        nameItemID = tonumber(nameItemID)
        if not TR.cache.items[nameItemID] then
            local item = Item:CreateFromItemID(nameItemID)
            item:ContinueOnItemLoad(function()
                local itemName, _, itemRarity = GetItemInfo(nameItemID)
                -- Debug("Create additional item for %d for %s", nameItemID, itemName)
                TR.cache.items[nameItemID] = { itemName = itemName }
            end)
        end
    end

    return TR.cache.items[itemID].textureID
end

function TrackResources:GetPosition()
    -- TODO: what was the bug in dungeons?
    local uiMapID = C_Map.GetBestMapForUnit("player");
    if not uiMapID then return end

    local position = C_Map.GetPlayerMapPosition(uiMapID, "player");
    if not position then return end

    return uiMapID, position:GetXY();
end

function TrackResources:NearExistingCloud(itemID, uiMapID, x, y)
    local playerX = x*100
    local playerY = y*100

    for posKey, itemData in pairs(TrackResourcesCfg.db[uiMapID][itemID]) do
        local dbX, dbY = string.match(posKey, "([0-9]+), ([0-9]+)")
        local distance = ((dbX - playerX) ^ 2 + (dbY - playerY) ^ 2) ^ 0.5
        if distance < TrackResourcesCfg.settings.minDistanceToCloud then return true end
    end

    return false
end

function TrackResources:AddResourceToDB(itemID, textureID, uiMapID, x, y, nameItemID)
    local posKey = string.format("%d, %d", math.floor((x*100)+0.5), math.floor((y*100)+0.5))

    if not TrackResourcesCfg.db[uiMapID] then TrackResourcesCfg.db[uiMapID] = {} end
    if not TrackResourcesCfg.db[uiMapID][itemID] then TrackResourcesCfg.db[uiMapID][itemID] = {} end
    if not TrackResourcesCfg.db[uiMapID][itemID][posKey] then
        if TR.motes[itemID] and TrackResources:NearExistingCloud(itemID, uiMapID, x, y) then return end
        local itemName = TR.cache.items[itemID].itemName
        if itemName and (nameItemID == nil or nameItemID > 0) then
            if nameItemID and nameItemID > 0 and TR.cache.items[nameItemID] then
                itemNameAs = TR.cache.items[nameItemID].itemName
                itemName = ("%s (%s)"):format(itemName, itemNameAs)
            end
            self:Print("Added %s (%s) (%0.2f, %0.2f)", itemName, GetRealZoneText(), x*100, y*100)
        end
        TrackResourcesCfg.db[uiMapID][itemID][posKey] = { itemID, textureID , uiMapID, x, y }
        TrackResourcesMap:AddIcon(uiMapID, posKey, TrackResourcesCfg.db[uiMapID][itemID][posKey])
        TrackResourcesMinimap:AddIcon(TrackResourcesCfg.db[uiMapID][itemID][posKey], self:GetPlayerWorldPosition())
    end
end

function TrackResources:GetPlayerWorldPosition()
    local pX, pY = UnitPosition("player")
    local vector = CreateVector2D(pX, pY)
    return vector
end

function TrackResources:Print(str, ...)
    if ... then str = str:format(...) end
    DEFAULT_CHAT_FRAME:AddMessage(("|cff6666ccTrackResources|r: %s"):format(str));
end

function TrackResources:Important(str, ...)
    if ... then str = str:format(...) end
    DEFAULT_CHAT_FRAME:AddMessage(("|cffdddd00TrackResources|r: %s"):format(str));
end

function TrackResources:typeValue(obj)
    local typeStr = type(obj)
    local valueStr = obj or "nil"

    return typeStr .. '/' .. valueStr
end

function Debug (str, ...)
    if ... then str = str:format(...) end
    DEFAULT_CHAT_FRAME:AddMessage(("|cffff0000TrackResources|r: %s"):format(str));
end

-- TODO: VALIDATE
function TrackResources:ParseCommands(text, chatBoxFrame)
    local _, _, cmd, args = string.find(text, "%s?(%w+)%s?(.*)")
    local commands = {
        zonemap = function() TrackResourcesMap:Toggle() end,
        minimap = function() TrackResourcesMinimap:Toggle() end,
        zonesize = function(size)
            if not size then
                self:PrintHelp()
                return
            end
            TrackResourcesCfg.settings.sizeZonemap = size
            TrackResourcesMap:UpdateIcons()
        end,
        minisize = function(size)
            if not size then
                self:PrintHelp()
                return
            end
            TrackResourcesCfg.settings.sizeMinimap = size
            TrackResourcesMinimap:UpdateIconSizes(size)
        end,
        alpha = function(alpha)
            TrackResourcesCfg.settings.alpha = tonumber(alpha)
            TrackResourcesMap:UpdateIcons()
        end,
        autocast = function()
            if TrackResourcesCfg.settings.autocast then
                TrackResourcesCfg.settings.autocast = false
                self:Print("Turned off casting tracking after death.")
            else
                TrackResourcesCfg.settings.autocast = true
                self:Print("Turned on casting tracking after death.")
            end
        end,
        prof = function()
            TrackResourcesCfg.settings.alwaysShow = not TrackResourcesCfg.settings.alwaysShow
            if TrackResourcesCfg.settings.alwaysShow then
                TR.isHerbalist = true
                TR.isMiner = true
                TR.isEngineer = true
                self:Print("Showing all resources on maps")
            else
                TrackResources:CheckProfessionSpells()
                if TR.isHerbalist then self:Print("Herbalism detected, showing herbs on maps") end
                if TR.isMiner then self:Print("Mining detected, showing ores on maps") end
                if TR.isEngineer then self:Print("Engineering detected, showing gas clouds on maps") end
            end
            TrackResourcesMinimap:RefreshIcons()
        end,
        filtermode = function(filterMode)
            filterMode = filterMode:upper()
            if filterMode == "LINEAR" or filterMode == "TRILINEAR" or filterMode == "NEAREST" then
                self:Print("Using texture filterMode %s", filterMode)
                TrackResourcesCfg.settings.filterMode = filterMode
                TrackResourcesMap:ResetIconCache()
                TrackResourcesMinimap:ReloadIcons(true)
            else
                self:PrintHelp()
            end
        end,
    }

    if commands[cmd] then
        commands[cmd](args)
    else
        self:PrintHelp()
    end
end

function TrackResources:PrintHelp()
    local function help(str)
        DEFAULT_CHAT_FRAME:AddMessage(("|cffcc66ccTrackResources|r: /tr %s"):format(str));
    end
    DEFAULT_CHAT_FRAME:AddMessage("|cffff66ffTrackResources|r: Help:")
    help("zonemap (toggles visibility of icons on the Zone map)")
    help("minimap (toggles visibility of icons on the Minimap)")
    help("zonesize x (size of icons on the Zone map, default 12)")
    help("minisize x (size of icons the Minimap, default 12)")
    help("autocast (toggle auto cast of gathering after death")
    help("alpha n.n (alpha/transparency level, from 0.0 to 1.0, default 0.75)")
    help("prof (toggles showing all profession resources, or currently learned professions)")
    help("filtermode x (type of texture filtering: LINEAR (default), NEAREST, TRILINEAR)")
end